/* $Id: lbunzip2.c,v 1.11.2.9 2008-08-19 15:35:08 lacos Exp $ */

#include <stdlib.h>       /* malloc() */
#include <limits.h>       /* CHAR_BIT */
#include <stdio.h>        /* flockfile() */
#include <unistd.h>       /* _exit() */
#include <errno.h>        /* errno */
#include <string.h>       /* strerror() */
#include <assert.h>       /* assert() */
#include <pthread.h>      /* pthread_mutex_t */
#include <inttypes.h>     /* uint64_t */

#include <bzlib.h>        /* BZ2_bzDecompressInit */

#include "lacos_rbtree.h" /* struct lacos_rbtree_node */
#include "lbunzip2.h"     /* struct splitq */


/* Bzip2 stream header and block header together. */
static const char unsigned intro[] = {
    0x42u, 0x5Au, 0x68u, 0x39u, 0x31u, 0x41u, 0x59u, 0x26u, 0x53u, 0x59
};

/* Bzip2 end of stream marker. */
static const char unsigned eos[] = {
    0x17u, 0x72u, 0x45u, 0x38u, 0x50u, 0x90u
};


static const char *pname;
static struct splitq splitq;
static struct muxq muxq;


static void
fail(const char *op, int err)
{
  /* protect strerror() by locking stderr */
  flockfile(stderr);
  (void)fprintf(stderr, "%s: %s: %s\n", pname, op, strerror(err));
  funlockfile(stderr);
  /* stderr is never fully buffered originally */
  _exit(EXIT_FAILURE);
}


static void *
xmalloc(size_t size)
{
  void *ret = malloc(size);

  if (0 == ret) {
    fail("malloc()", errno);
  }

  return ret;
}


static long
xstrtol(char *str, long lower, long upper)
{
  long tmp;
  char *endptr;

  errno = 0;
  tmp = strtol(str, &endptr, 10);
  if ('\0' == *str || '\0' != *endptr || 0 != errno
      || tmp < lower || tmp > upper) {
    (void)fprintf(stderr, "%s: failed to parse \"%s\" as a long in"
        " [%ld..%ld]\n", pname, str, lower, upper);
    _exit(EXIT_FAILURE);
  }

  return tmp;
}


static void
xmutex_init(pthread_mutex_t *mutex)
{
  int ret;
  pthread_mutexattr_t attr;

  ret = pthread_mutexattr_init(&attr);
  if (0 != ret) {
    fail("pthread_mutexattr_init()", ret);
  }

  ret = pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK);
  if (0 != ret) {
    fail("pthread_mutexattr_settype()", ret);
  }

  ret = pthread_mutex_init(mutex, &attr);
  if (0 != ret) {
    fail("pthread_mutex_init()", ret);
  }

  ret = pthread_mutexattr_destroy(&attr);
  if (0 != ret) {
    fail("pthread_mutexattr_destroy()", ret);
  }
}


static void
xmutex_lock(pthread_mutex_t *mutex)
{
  int ret = pthread_mutex_lock(mutex);

  if (0 != ret) {
    fail("pthread_mutex_lock()", ret);
  }
}


static void
xmutex_unlock(pthread_mutex_t *mutex)
{
  int ret = pthread_mutex_unlock(mutex);

  if (0 != ret) {
    fail("pthread_mutex_unlock()", ret);
  }
}


static void
xmutex_destroy(pthread_mutex_t *mutex)
{
  int ret = pthread_mutex_destroy(mutex);

  if (0 != ret) {
    fail("pthread_mutex_destroy()", ret);
  }
}


static void
xcond_init(pthread_cond_t *cond)
{
  int ret = pthread_cond_init(cond, 0);
  if (0 != ret) {
    fail("pthread_cond_init()", ret);
  }
}


static void
xcond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex)
{
  int ret = pthread_cond_wait(cond, mutex);

  if (0 != ret) {
    fail("pthread_cond_wait()", ret);
  }
}


static void
xcond_signal(pthread_cond_t *cond)
{
  int ret = pthread_cond_signal(cond);

  if (0 != ret) {
    fail("pthread_cond_signal()", ret);
  }
}


static void
xcond_broadcast(pthread_cond_t *cond)
{
  int ret = pthread_cond_broadcast(cond);

  if (0 != ret) {
    fail("pthread_cond_broadcast()", ret);
  }
}


static void
xcond_destroy(pthread_cond_t *cond)
{
  int ret = pthread_cond_destroy(cond);

  if (0 != ret) {
    fail("pthread_cond_destroy()", ret);
  }
}


static void
xcreate(pthread_t *thread, void *(*routine)(void *))
{
  int ret;

  ret = pthread_create(thread, 0, routine, 0);
  if (0 != ret) {
    fail("pthread_create()", ret);
  }
}


static void
xjoin(pthread_t thread)
{
  int ret;

  ret = pthread_join(thread, 0);
  if (0 != ret) {
    fail("pthread_join()", ret);
  }
}


static int
muxb_cmp(const void *muxb_new, const void *muxb)
{
  const struct muxb *m_new = muxb_new,
      *m = muxb;

  return m_new->serial < m->serial ? -1
      : m_new->serial > m->serial ? 1
      : m_new->subserial < m->subserial ? -1
      : m_new->subserial > m->subserial ? 1
      : 0;
}


static void *
muxq_alloc(size_t size, void *ignored_ctl)
{
  return malloc(size);
}


static void
muxq_dealloc(void *ptr, void *ignored_ctl)
{
  free(ptr);
}


static void
mux(void)
{
  for (;;) {
    struct lacos_rbtree_node *head;
    struct muxb *muxb;

    /* Grab a sub-block to write out. */
    xmutex_lock(&muxq.lock);

    for (;;) {
      ++muxq.c_avail_or_exit;

      head = lacos_rbtree_min(muxq.order);
      if (0 != head) {
        muxb = *(void **)head;

        if (muxq.next_serial == muxb->serial
            && muxq.next_subserial == muxb->subserial) {
          break;
        }
      }
      else {
        if (0u == muxq.num_workers) {
          break;
        }
      }

      ++muxq.w_avail_or_exit;
      xcond_wait(&muxq.avail_or_exit, &muxq.lock);
    }

    if (0 == head) {
      /* "order" is empty and all workers exited */
      xmutex_unlock(&muxq.lock);
      break;
    }

    if (0 == muxb->source) {
      ++muxq.next_subserial;
    }
    else {
      ++muxq.next_serial;
      muxq.next_subserial = 0lu;
    }

    lacos_rbtree_delete(
        &muxq.order,  /* new_root */
        head,         /* old_node */
        0,            /* old_data */
        muxq_dealloc, /* dealloc() */
        0             /* alloc_ctl */
    );

    xmutex_unlock(&muxq.lock);


    /* Write out "muxb". */
    {
      char unsigned *dp;

      dp = muxb->decompr;
      while (muxb->used > 0u) {
        ssize_t written;

        written = write(STDOUT_FILENO, dp, muxb->used > (size_t)SSIZE_MAX
            ? (size_t)SSIZE_MAX : muxb->used);
        if (-1 == written) {
          fail("write()", errno);
        }

        muxb->used -= (size_t)written;
        dp += written;
      }
    }

    /* Return source "splitb" to the splitter if last subserial processed. */
    if (0 != muxb->source) {
      xmutex_lock(&splitq.free.lock);
      muxb->source->next = 0;
      if (0 == splitq.free.tail) {
        splitq.free.head = muxb->source;
        xcond_signal(&splitq.free.avail);
      }
      else {
        splitq.free.tail->next = muxb->source;
      }
      splitq.free.tail = muxb->source;
      xmutex_unlock(&splitq.free.lock);
    }

    /* Release "muxb". */
    free(muxb);
  }
}


static void
work_compl(struct splitb *splitb)
{
  char unsigned eos_crc[sizeof eos + NUM_CRC];
  unsigned fb,
      ub;

  assert(splitb->num_fullbytes <= sizeof splitb->bzip2);

  if (splitb->num_fullbytes < sizeof intro + NUM_CRC + sizeof eos) {
    (void)fprintf(stderr, "%s: input block too short\n", pname);
    _exit(EXIT_FAILURE);
  }

  fb = splitb->num_freebits;
  ub = CHAR_BIT - fb;

  splitb->num_fullbytes -= sizeof eos;
  if (0u < ub) {
    splitb->incomplete = (unsigned)splitb->bzip2[splitb->num_fullbytes] >> fb;
  }

  if (sizeof splitb->bzip2 - splitb->num_fullbytes
      < sizeof eos_crc + (0u < ub)) {
    (void)fprintf(stderr, "%s: input block too long\n", pname);
    _exit(EXIT_FAILURE);
  }

  if (0u < ub) {
    size_t ctr;

    (void)memcpy(eos_crc, eos, sizeof eos);
    (void)memcpy(eos_crc + sizeof eos, splitb->bzip2 + sizeof intro, NUM_CRC);

    splitb->bzip2[splitb->num_fullbytes++]
        = splitb->incomplete << fb | (unsigned)eos_crc[0u] >> ub;
    for (ctr = 1u; ctr < sizeof eos_crc; ++ctr) {
      splitb->bzip2[splitb->num_fullbytes++]
          = (unsigned)eos_crc[ctr - 1u] << fb | (unsigned)eos_crc[ctr] >> ub;
    }
    splitb->bzip2[splitb->num_fullbytes++]
        = (unsigned)eos_crc[sizeof eos_crc - 1u] << fb;
  }
  else {
    (void)memcpy(splitb->bzip2 + splitb->num_fullbytes, eos, sizeof eos);
    splitb->num_fullbytes += sizeof eos;
    (void)memcpy(splitb->bzip2 + splitb->num_fullbytes,
        splitb->bzip2 + sizeof intro, NUM_CRC);
    splitb->num_fullbytes += NUM_CRC;
  }
}


static void
work_decompr(struct splitb *splitb)
{
  long unsigned subserial;
  bz_stream strm;
  int bzret;

  subserial = 0lu;
  strm.next_in = (char *)splitb->bzip2;
  strm.avail_in = splitb->num_fullbytes;
  strm.next_out = 0; /* safety */
  strm.avail_out = 0u; /* safety */
  strm.bzalloc = 0;
  strm.bzfree = 0;
  strm.opaque = 0;
  if (BZ_OK != BZ2_bzDecompressInit(&strm,
      /* verbosity */ 0, /* small */ 0)) {
    (void)fprintf(stderr, "%s: BZ2_bzDecompressInit() failed\n", pname);
    _exit(EXIT_FAILURE);
  }

  do {
    struct muxb *muxb;
    struct lacos_rbtree_node *new_node;

    muxb = xmalloc(sizeof *muxb);
    strm.next_out = (char *)muxb->decompr;
    strm.avail_out = sizeof muxb->decompr;
    bzret = BZ2_bzDecompress(&strm);

#define CASE(x) \
      case x: \
        do { \
          (void)fprintf(stderr, "%s: BZ2_bzDecompress(): " #x "\n", \
              pname); \
          _exit(EXIT_FAILURE); \
        } while (0)

    switch(bzret) {
      CASE(BZ_PARAM_ERROR);
      CASE(BZ_DATA_ERROR);
      CASE(BZ_DATA_ERROR_MAGIC);
      CASE(BZ_MEM_ERROR);

      case BZ_STREAM_END:
        /* The splitter finds real block headers and EOS markers for sure. */
        assert(0u == strm.avail_in);
        break;

      case BZ_OK:
        /* If what's needed can't be output space, we fail. */
        if (0u < strm.avail_out) {
          (void)fprintf(stderr, "%s: the splitter misrecognized a bit-sequence"
              " as a block delimiter\n", pname);
          _exit(EXIT_FAILURE);
        }
        break;

      default:
        (void)fprintf(stderr, "%s: BZ2_bzDecompress(): unkown return"
            " value\n", pname);
        _exit(EXIT_FAILURE);
    }

#undef CASE

    muxb->serial = splitb->serial;
    muxb->subserial = subserial++;
    muxb->source = BZ_STREAM_END == bzret ? splitb : 0;
    muxb->used = sizeof muxb->decompr - (size_t)strm.avail_out;

    /* Push sub-block to muxer. */
    xmutex_lock(&muxq.lock);
    if (-1 == lacos_rbtree_insert(
        &muxq.order, /* new_root */
        &new_node,   /* new_node */
        muxb,        /* new_data */
        muxb_cmp,    /* cmp() */
        muxq_alloc,  /* alloc() */
        0            /* alloc_ctl */
    )) {
      /* (serial, subserial) pair collision shouldn't happen */
      assert(0 == new_node);
      (void)fprintf(stderr, "%s: lacos_rbtree_insert(): out of memory\n",
          pname);
      _exit(EXIT_FAILURE);
    }
    if (muxq.next_serial == muxb->serial
        && muxq.next_subserial == muxb->subserial) {
      xcond_signal(&muxq.avail_or_exit);
    }
    xmutex_unlock(&muxq.lock);
  } while (BZ_OK == bzret);

  if (BZ_OK != BZ2_bzDecompressEnd(&strm)) {
      (void)fprintf(stderr, "%s: BZ2_bzDecompressEnd() failed\n", pname);
    _exit(EXIT_FAILURE);
  }
}


static void *
work(void *ignored)
{
  for (;;) {
    struct splitb *splitb;

    /* Grab a block to work on. */
    xmutex_lock(&splitq.used.lock);
    for (;;) {
      ++splitq.used.c_avail_or_eof;
      if (0 != splitq.used.head || 0 != splitq.used.eof) {
        break;
      }
      ++splitq.used.w_avail_or_eof;
      xcond_wait(&splitq.used.avail_or_eof, &splitq.used.lock);
    }
    if (0 == splitq.used.head) {
      /* No blocks available and splitter exited. */
      xmutex_unlock(&splitq.used.lock);
      break;
    }
    splitb = splitq.used.head;
    splitq.used.head = splitb->next;
    if (0 == splitq.used.head) {
      splitq.used.tail = 0;
    }
    xmutex_unlock(&splitq.used.lock);

    /* Replace trailing bzip2 block header or EOS with EOS, CRC, padding. */
    work_compl(splitb);

    /* Decompress. */
    work_decompr(splitb);
  }

  /* Notify muxer if last worker exits. */
  xmutex_lock(&muxq.lock);
  --muxq.num_workers;
  if (0u == muxq.num_workers) {
    xcond_signal(&muxq.avail_or_exit);
  }
  xmutex_unlock(&muxq.lock);

  return 0;
}


static void *
split(void *ignored)
{
  long unsigned serial;
  uint64_t search;
  struct splitb *splitb;

  serial = 0lu;
  search = 0u;
  splitb = 0;

  flockfile(stdin);
  for (;;) {
    int c;
    unsigned i;

    c = getchar_unlocked();
    if (EOF == c) {
      if (ferror(stdin)) {
        fail("getchar_unlocked()", errno);
      }
      break;
    }

    i = CHAR_BIT;
    do {
      unsigned bit;
      int is_hdr;

      bit = c >> --i & 1u;
      search = (search << 1 | bit) & MAGIC_MASK;
      is_hdr = MAGIC_HDR == search;

      if (0 != splitb) {
        int is_end;

        /* Push bit. */
        {
          splitb->incomplete = splitb->incomplete << 1 | bit;
          if (0u == --splitb->num_freebits) {
            if (sizeof splitb->bzip2 == splitb->num_fullbytes) {
              (void)fprintf(stderr, "%s: input block too long\n", pname);
              _exit(EXIT_FAILURE);
            }
            splitb->bzip2[splitb->num_fullbytes++] = splitb->incomplete;
            splitb->num_freebits = CHAR_BIT;
          }
        }

        is_end = MAGIC_EOS == search;
        if (is_end || is_hdr) {
          /* Push used splitb. */
          {
            xmutex_lock(&splitq.used.lock);
            if (0 == splitq.used.tail) {
              splitq.used.head = splitb;
              xcond_broadcast(&splitq.used.avail_or_eof);
            }
            else {
              splitq.used.tail->next = splitb;
            }
            splitq.used.tail = splitb;
            xmutex_unlock(&splitq.used.lock);
          }

          if (is_end) {
            splitb = 0;
          }
        } /* Bzip2 block header or end of stream reached. */
      } /* A splitb was available when entering. */


      if (is_hdr) {
        /* Grab a free splitb to fill. */
        {
          xmutex_lock(&splitq.free.lock);
          for (;;) {
            ++splitq.free.c_avail;
            if (0 != splitq.free.head) {
              break;
            }
            ++splitq.free.w_avail;
            xcond_wait(&splitq.free.avail, &splitq.free.lock);
          }
          splitb = splitq.free.head;
          splitq.free.head = splitb->next;
          if (0 == splitq.free.head) {
            splitq.free.tail = 0;
          }
          xmutex_unlock(&splitq.free.lock);
        }

        /* Initialize splitb. */
        {
          splitb->next = 0;
          splitb->serial = serial++;
          splitb->num_fullbytes = sizeof intro;
          (void)memcpy(splitb->bzip2, intro, sizeof intro);
          splitb->incomplete = 0u;
          splitb->num_freebits = CHAR_BIT;
        }
      } /* New bzip2 block started. */
    } while (i > 0u);
  }
  funlockfile(stdin);

  if (0 != splitb) {
    (void)fprintf(stderr, "%s: premature EOF on stdin\n", pname);
    _exit(EXIT_FAILURE);
  };

  /* Notify the workers about the EOF. */
  xmutex_lock(&splitq.used.lock);
  splitq.used.eof = 1;
  xcond_broadcast(&splitq.used.avail_or_eof);
  xmutex_unlock(&splitq.used.lock);

  return 0;
}


static void
usage(void)
{
  (void)fprintf(stderr,
      "%1$s-%2$s: usage: %1$s WORKER-THREADS\n"
      "Parallel bunzip2 filter. WORKER-THREADS must be in [%3$u..%4$u].\n"
      "Copyright (C) 2008 Laszlo Ersek. Released under the GNU GPLv2+.\n",
      pname, VERSION, 1u, MX_WORKER);
  _exit(EXIT_FAILURE);
}


static void
init_splitq(unsigned num_splitbs)
{
  unsigned i;

  splitq.lists_base = xmalloc((size_t)num_splitbs * sizeof *splitq.lists_base);
  for (i = 0u; i < num_splitbs; ++i) {
    splitq.lists_base[i].next = &splitq.lists_base[i + 1u];
  }

  xmutex_init(&splitq.used.lock);
  splitq.used.head = 0;
  splitq.used.tail = 0;
  splitq.used.eof = 0;
  xcond_init(&splitq.used.avail_or_eof);
  splitq.used.c_avail_or_eof = 0lu;
  splitq.used.w_avail_or_eof = 0lu;

  xmutex_init(&splitq.free.lock);
  splitq.free.head = splitq.lists_base;
  splitq.free.tail = splitq.lists_base + (num_splitbs - 1u);
  splitq.free.tail->next = 0;
  xcond_init(&splitq.free.avail);
  splitq.free.c_avail = 0lu;
  splitq.free.w_avail = 0lu;
}


static void
uninit_splitq(void)
{
  xcond_destroy(&splitq.free.avail);
  assert(0 != splitq.free.tail);
  assert(0 == splitq.free.tail->next);
  assert(0 != splitq.free.head);
  xmutex_destroy(&splitq.free.lock);

  xcond_destroy(&splitq.used.avail_or_eof);
  assert(0 != splitq.used.eof);
  assert(0 == splitq.used.tail);
  assert(0 == splitq.used.head);
  xmutex_destroy(&splitq.used.lock);

  free(splitq.lists_base);
}


static void
init_muxq(unsigned num_workers)
{
  xmutex_init(&muxq.lock);
  muxq.order = 0;
  xcond_init(&muxq.avail_or_exit);
  muxq.c_avail_or_exit = 0lu;
  muxq.w_avail_or_exit = 0lu;
  muxq.num_workers = num_workers;
  muxq.next_serial = 0lu;
  muxq.next_subserial = 0lu;
}


static void
uninit_muxq(void)
{
  assert(0lu == muxq.next_subserial);
  assert(0u == muxq.num_workers);
  xcond_destroy(&muxq.avail_or_exit);
  assert(0 == muxq.order);
  xmutex_destroy(&muxq.lock);
}


static void
print_stats(void)
{
  (void)fprintf(stderr,
      "%1$s: the number of times:\n"
      "%1$s: - the splitter tried to get a free splitb:    %3$*2$lu\n"
      "%1$s: - the splitter had to wait for a free splitb: %4$*2$lu\n"
      "%1$s: - any worker tried to get a used splitb:      %5$*2$lu\n"
      "%1$s: - any worker had to wait for a used splitb:   %6$*2$lu\n"
      "%1$s: - the muxer tried to get the next muxb:       %7$*2$lu\n"
      "%1$s: - the muxer had to wait for the next muxb:    %8$*2$lu\n", pname,
      (int)sizeof(long unsigned) * (int)CHAR_BIT / 3 + 1,
      splitq.free.c_avail,
      splitq.free.w_avail,
      splitq.used.c_avail_or_eof,
      splitq.used.w_avail_or_eof,
      muxq.c_avail_or_exit,
      muxq.w_avail_or_exit
  );
}


int
main(int argc, char **argv)
{
  unsigned num_workers,
      i;
  pthread_t splitter,
      *worker;

  pname = strrchr(argv[0], '/');
  pname = pname ? pname + 1 : argv[0];

  if (2 != argc) {
    usage();
  }

  num_workers = xstrtol(argv[1], 1, MX_WORKER);

  init_splitq(num_workers * FAC_SPLITQ);
  init_muxq(num_workers);
  worker = xmalloc((size_t)num_workers * sizeof *worker);

  xcreate(&splitter, split);
  for (i = 0u; i < num_workers; ++i) {
    xcreate(&worker[i], work);
  }

  mux();

  i = num_workers;
  do {
    xjoin(worker[--i]);
  } while (i > 0u);
  xjoin(splitter);

  free(worker);
  print_stats();
  uninit_muxq();
  uninit_splitq();

  return EXIT_SUCCESS;
}
